-- Top-level design for using framework firmware with KU115 on HTG K800 board
--
-- Raghunandan Shukla (TIFR), Kristian Harder (RAL), Tom Williams (RAL)
-- based on code from Dave Newbold

library IEEE;
use IEEE.std_logic_1164.all;

-- IPBus
use work.ipbus.all;

-- EMP constants and data types 
use work.emp_ttc_decl.all;
use work.emp_data_types.all;

use work.ipbus_decode_top_emp.all;
use work.emp_device_decl.all;
use work.emp_slim_top_decl;

library UNISIM;
use UNISIM.VComponents.all;


entity top is
  generic (
    UNIX_TIME             : std_logic_vector(31 downto 0) := X"00000000";
    GIT_REPOS_NAME        : std_logic_vector              := X"";
    GIT_REPOS_SHA         : std_logic_vector              := X"";
    GIT_REPOS_CLEAN       : std_logic_vector              := X"";
    GIT_REPOS_REF         : std_logic_vector              := X"";
    GITLAB_CI_PROJECT_ID  : integer                       := 0;
    GITLAB_CI_PIPELINE_ID : integer                       := 0;
    GITLAB_CI_JOB_ID      : integer                       := 0
    );
  port(
    -- PCIe clock and reset
    pcie_sys_clk_p : in  std_logic;
    pcie_sys_clk_n : in  std_logic;
    pcie_sys_rst   : in  std_logic;
    -- PCIe lanes
    pcie_rx_p      : in  std_logic_vector(0 downto 0);
    pcie_rx_n      : in  std_logic_vector(0 downto 0);
    pcie_tx_p      : out std_logic_vector(0 downto 0);
    pcie_tx_n      : out std_logic_vector(0 downto 0);
    -- External oscillator
    clk125         : in  std_logic;
    -- Legacy TTC interface
    ttc_clk_p      : in  std_logic;
    ttc_clk_n      : in  std_logic;
    ttc_rx_p       : in  std_logic;
    ttc_rx_n       : in  std_logic;
    -- TCDS2 interface
    tcds_clk_p     : in  std_logic;
    tcds_clk_n     : in  std_logic;
    tcds_rx_p      : in  std_logic;
    tcds_rx_n      : in  std_logic;
    tcds_tx_p      : out std_logic;
    tcds_tx_n      : out std_logic;
    -- 10g eth signals
    gbe_rx_p          : in std_logic_vector(1 downto 0);
    gbe_rx_n          : in std_logic_vector(1 downto 0);
    gbe_tx_p          : out std_logic_vector(1 downto 0);
    gbe_tx_n          : out std_logic_vector(1 downto 0);
    gbe_mgt_refclk_p  : in std_logic;
    gbe_mgt_refclk_n  : in std_logic;
    -- Transceiver ref-clocks
    refclkp        : in  std_logic_vector(N_REFCLK - 1 downto 0);
    refclkn        : in  std_logic_vector(N_REFCLK - 1 downto 0);
    -- Heartbeat LED
    heartbeat_led  : out std_logic
    );

end top;

architecture rtl of top is

  signal pcie_sys_rst_n : std_logic;

  -- IPBus signals
  signal ipb_clk, ipb_rst : std_logic;

  signal ipb_w : ipb_wbus;
  signal ipb_r : ipb_rbus;

  signal ipb_w_array : ipb_wbus_array(N_SLAVES - 1 downto 0);
  signal ipb_r_array : ipb_rbus_array(N_SLAVES - 1 downto 0);

  -- LHC clocks and signals
  signal clk40ish : std_logic;

  -- Clock domain reset outputs
  signal clk40    : std_logic;
  signal rst40    : std_logic;
  signal clk_p    : std_logic;
  signal rst_p    : std_logic;
  signal clks_aux : std_logic_vector(2 downto 0);
  signal rsts_aux : std_logic_vector(2 downto 0);

  -- TTC signals  
  signal ttc_l1a, ttc_l1a_dist, dist_lock, oc_flag, ec_flag, ttc_l1a_throttle              : std_logic;
  signal ttc_cmd, ttc_cmd_dist                                                             : ttc_cmd_t;
  signal bunch_ctr                                                                         : bctr_t;
  signal orb_ctr                                                                           : eoctr_t;

  -- Others
  signal nuke, soft_rst : std_logic;

  -- Datapath signals
  signal payload_d, payload_q : ldata(N_REGION * 4 - 1 downto 0);
  signal clkmon               : std_logic_vector(3 downto 0);
  --signal cap_bus                  : daq_cap_bus;
  --signal daq_bus_top, daq_bus_bot : daq_bus;
  signal ctrs                 : ttc_stuff_array(N_REGION - 1 downto 0);
  signal rst_loc, clken_loc   : std_logic_vector(N_REGION - 1 downto 0);

  signal clk125_s             : std_logic;

  -- 10g
  signal gbe_ldata : ldata(1 downto 0);
  signal gbe_lff : std_logic_vector(1 downto 0);
  
  -- KH tmp
  signal clk_40_div, clk_p_div : std_logic;


begin

-- Infrastructure

  pcie_rst_passthrough : if emp_slim_top_decl.PCIE_RST_ACTIVE_LEVEL = '0' generate
    pcie_sys_rst_n <= pcie_sys_rst;
  end generate;

  pcie_rst_invert : if emp_slim_top_decl.PCIE_RST_ACTIVE_LEVEL = '1' generate
    pcie_sys_rst_n <= not pcie_sys_rst;
  end generate;

  infra : entity work.emp_infra_pcie
    port map(
      pcie_sys_clk_p => pcie_sys_clk_p,
      pcie_sys_clk_n => pcie_sys_clk_n,
      pcie_sys_rst_n => pcie_sys_rst_n,
      pcie_rx_p      => pcie_rx_p,
      pcie_rx_n      => pcie_rx_n,
      pcie_tx_p      => pcie_tx_p,
      pcie_tx_n      => pcie_tx_n,
      nuke           => nuke,
      soft_rst       => soft_rst,
      ipb_clk        => ipb_clk,
      ipb_rst        => ipb_rst,
      ipb_in         => ipb_r,
      ipb_out        => ipb_w,
      clk_40ish      => clk40ish,
      onehz          => heartbeat_led
      );

-- ipbus fabric selector

  fabric : entity work.ipbus_fabric_sel
    generic map(
      NSLV      => N_SLAVES,
      SEL_WIDTH => IPBUS_SEL_WIDTH)
    port map(
      ipb_in          => ipb_w,
      ipb_out         => ipb_r,
      sel             => ipbus_sel_top_emp(ipb_w.ipb_addr),
      ipb_to_slaves   => ipb_w_array,
      ipb_from_slaves => ipb_r_array
      );


-- info block (constant registers containing version numbers, build info, etc)
  info : entity work.emp_info
    generic map (
      UNIX_TIME             => UNIX_TIME,
      GIT_REPOS_NAME        => GIT_REPOS_NAME,
      GIT_REPOS_SHA         => GIT_REPOS_SHA,
      GIT_REPOS_CLEAN       => GIT_REPOS_CLEAN,
      GIT_REPOS_REF         => GIT_REPOS_REF,
      GITLAB_CI_PROJECT_ID  => GITLAB_CI_PROJECT_ID,
      GITLAB_CI_PIPELINE_ID => GITLAB_CI_PIPELINE_ID,
      GITLAB_CI_JOB_ID      => GITLAB_CI_JOB_ID
      )
    port map(
      clk     => ipb_clk,
      rst     => ipb_rst,
      ipb_in  => ipb_w_array(N_SLV_INFO),
      ipb_out => ipb_r_array(N_SLV_INFO)
      );

  ctrl : entity work.emp_ctrl
    port map (
      clk      => ipb_clk,
      rst      => ipb_rst,
      ipb_in   => ipb_w_array(N_SLV_CTRL),
      ipb_out  => ipb_r_array(N_SLV_CTRL),
      nuke     => nuke,
      soft_rst => soft_rst,
      debug    => (others => '0')
      );

-- TTC block
  ttc : entity work.emp_ttc
    generic map (
      ENABLE_LEGACY_TTC => true,
      ENABLE_TCDS2      => true,
      EXT_CLK_FREQ      => 40.0
      )
    port map(
      clk_ipb      => ipb_clk,
      rst_ipb      => ipb_rst,
      ipb_in       => ipb_w_array(N_SLV_TTC),
      ipb_out      => ipb_r_array(N_SLV_TTC),
      clk125       => clk125,
      clk40ish_in  => clk40ish,
      clk40        => clk40,
      rst40        => rst40,
      clk_p        => clk_p,
      rst_p        => rst_p,
      clks_aux     => clks_aux,
      rsts_aux     => rsts_aux,
      ttc_clk_p    => ttc_clk_p,
      ttc_clk_n    => ttc_clk_n,
      ttc_rx_p     => ttc_rx_p,
      ttc_rx_n     => ttc_rx_n,
      tcds_clk_p   => tcds_clk_p,
      tcds_clk_n   => tcds_clk_n,
      tcds_rx_p    => tcds_rx_p,
      tcds_rx_n    => tcds_rx_n,
      tcds_tx_p    => tcds_tx_p,
      tcds_tx_n    => tcds_tx_n,
      ttc_cmd      => ttc_cmd,
      ttc_cmd_dist => ttc_cmd_dist,
      ttc_l1a      => ttc_l1a,
      ttc_l1a_dist => ttc_l1a_dist,
      dist_lock    => dist_lock,
      bunch_ctr    => bunch_ctr,
      evt_ctr      => (others => '0'),
      orb_ctr      => orb_ctr,
      oc_flag      => oc_flag,
      ec_flag      => ec_flag,
      monclk       => clkmon,
      clk125_o     => clk125_s
      );

-- Datapath block
  datapath : entity work.emp_datapath
    port map(
      clk         => ipb_clk,
      rst         => ipb_rst,
      ipb_in      => ipb_w_array(N_SLV_DATAPATH),
      ipb_out     => ipb_r_array(N_SLV_DATAPATH),
      board_id    => (others => '0'),
      clk125      => clk125_s,
      clk40       => clk40,
      clk_p       => clk_p,
      rst_p       => rst_p,
      ttc_cmd     => ttc_cmd_dist,
      ttc_l1a     => ttc_l1a_dist,
      lock        => dist_lock,
      ctrs_out    => ctrs,
      rst_out     => rst_loc,
      clken_out   => clken_loc,
      refclkp     => refclkp,
      refclkn     => refclkn,
      clkmon      => clkmon,
      q           => payload_d,
      d           => payload_q
      );

  -- And finally, the payload
  -- KH ILA_DEBUG for payload defaults to false ...
  -- if desired set generic here and enable include the xci
  -- in the eth dep
  payload : entity work.emp_payload
    port map(
      clk         => ipb_clk,
      rst         => ipb_rst,
      ipb_in      => ipb_w_array(N_SLV_PAYLOAD),
      ipb_out     => ipb_r_array(N_SLV_PAYLOAD),
      clk40       => clk40,	        
      clk_payload => clks_aux,
      rst_payload => rsts_aux,
      clk_p       => clk_p,
      rst_loc     => rst_loc,
      clken_loc   => clken_loc,
      ctrs        => ctrs,
      bc0         => open,
      d           => payload_d,
      q           => payload_q,
      gpio        => open,
      gpio_en     => open,
      -- 10g
      gbe_q             => gbe_ldata,
      gbe_backpressure  => gbe_lff
      );


  --
  -- 10g
  --
  eth10g : entity work.emp_eth10g_interface
    generic map ( ETH10G_CHANNELS => 2 )
    port map (
      clk_ipb         => ipb_clk,
      rst_ipb         => ipb_rst,
      ipb_in          => ipb_w_array(N_SLV_ETH10G),
      ipb_out         => ipb_r_array(N_SLV_ETH10G),
      --
      clk_p           => clk_p,
      rst_p           => rst_p,
      clk125          => clk125,
      -- Input from payload
      ldata_in        => gbe_ldata,
      -- backpressure to payload
      lff             => gbe_lff,
      --10g
      gbe_rx_p          => gbe_rx_p,
      gbe_rx_n          => gbe_rx_n,
      gbe_tx_p          => gbe_tx_p,
      gbe_tx_n          => gbe_tx_n,
      gbe_mgt_refclk_p  => gbe_mgt_refclk_p,
      gbe_mgt_refclk_n  => gbe_mgt_refclk_n
      );
  
  


-- KH tmp to check clk_p
  div : entity work.freq_ctr_div
    generic map(
      N_CLK => 2
      )
    port map(
      clk(0)    => clk_p,
      clk(1)    => clk40,
      clkdiv(0) => clk_p_div,
      clkdiv(1) => clk_40_div
      );

-- Clock frequency monitor

  ctr : entity work.ipbus_freq_ctr
    generic map(
      N_CLK => 2
      )
    port map(
      clk                => ipb_clk,
      rst                => ipb_rst,
      ipb_in             => ipb_w_array(N_SLV_FREQ),
      ipb_out            => ipb_r_array(N_SLV_FREQ),
      clkdiv(0)          => clk_p_div,
      clkdiv(1)          => clk_40_div
      );



end rtl;

