library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use work.emp_data_types.all;

use work.ipbus.all;
use work.ipbus_reg_types.all;


entity LinkAggregatorCore is
    generic (
        N_LINKS : integer
    );
    port (
        --- Input Ports ---
        clk_p              : in std_logic;
        reset              : in std_logic;
        packet_start       : in std_logic;
        links_in           : in ldata(N_LINKS - 1 downto 0);
        --- Output Ports ---
        links_out          : out ldata(7 downto 0) := (others => LWORD_NULL);
        uncompressed_stubs : out ldata(N_LINKS * 8 - 1 downto 0) := (others => LWORD_NULL);
        readout_reset      : out std_logic := '0';
        --- Debug Ports ---
        status_bits        : out std_logic_vector(15 downto 0);
        --- IPBus Ports ---
        non_empty_counts   : out ipb_reg_v(8*N_LINKS - 1 downto 0);
        bx_valid_count_out : out ipb_reg_v(0 downto 0)
    );
end LinkAggregatorCore;


architecture compressed of LinkAggregatorCore is

    constant cFIFOWidth : integer := 64;


    -- FIFO IP Coress

    COMPONENT link_agg_single_link_fifo
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        overflow     : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(3 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;
    COMPONENT link_agg_single_link_fifo_fwft
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        overflow     : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(4 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;
    COMPONENT link_agg_all_links_fifo
    PORT (
        --- Input Ports ---
        clk          : IN STD_LOGIC;
        srst         : IN STD_LOGIC;
        din          : IN STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        wr_en        : IN STD_LOGIC;
        rd_en        : IN STD_LOGIC;
        --- Output Ports ---
        dout         : OUT STD_LOGIC_VECTOR(cFIFOWidth - 1 DOWNTO 0);
        full         : OUT STD_LOGIC;
        overflow     : OUT STD_LOGIC;
        empty        : OUT STD_LOGIC;
        almost_empty : OUT STD_LOGIC;
        valid        : OUT STD_LOGIC;
        data_count   : OUT STD_LOGIC_VECTOR(5 DOWNTO 0);
        wr_rst_busy  : OUT STD_LOGIC;
        rd_rst_busy  : OUT STD_LOGIC
    );
    END COMPONENT;


    constant cBoxcarBx  : integer := 8;
    
    type tLinkDataArray is array(N_LINKS - 1 downto 0) of std_logic_vector(cFIFOWidth - 1 downto 0);
    type t2DDataArray is array(cBoxcarBx - 1 downto 0) of tLinkDataArray;
    type tEnableArray is array(cBoxcarBx - 1 downto 0) of std_logic_vector(N_LINKS - 1 downto 0);

    constant cNullEnableArray    : tEnableArray                           := (others => (others => '0'));
    constant cNullEnableArrayRow : std_logic_vector(N_LINKS - 1 downto 0) := (others => '0');
    
    signal data                  : t2DDataArray                           := (others => (others => (others => '0')));
    signal rd_en_route_array     : tEnableArray                           := (others => (others => '0'));
    signal valid_route_array     : tEnableArray                           := (others => (others => '0'));
    signal counter               : integer                                := 0;


    function compress64bStubto32b(stub: std_logic_vector) return std_logic_vector is
        variable output : std_logic_vector(31 downto 0);
    begin
        output(21 downto 0)  := stub(21 downto 0); -- Common stub format
        output(22)           := stub(50); -- CIC Index
        output(23)           := stub(49); -- 2S/PS
        output(27 downto 24) := stub(63 downto 60); -- Link ID
        output(31 downto 28) := (others => '0');
        return output;
    end function;

    signal output_overflow_array : std_logic_vector(cBoxcarBx - 1 downto 0) := (others => '0');
    signal route_overflow_array  : tEnableArray                             := cNullEnableArray;
    signal cache_overflow_array  : tEnableArray                             := cNullEnableArray;
    
    signal bx_valid_array        : std_logic_vector(N_LINKS * cBoxcarBx - 1 downto 0) := (others => '0');
    
    signal non_empty_fifo_count  : ipb_reg_v(8*N_LINKS - 1 downto 0)        := (others => (others => '0'));
    signal bx_valid_count        : ipb_reg_v(0 downto 0)                    := (others => (others => '0'));
    signal pBx_counter           : integer                                  := 0;

begin

    --==============================--
    genLinkCache : for i in N_LINKS - 1 downto 0 generate
    -- Generator to produce the caching and routing FIFOs for each link
    --==============================--
        signal wr_en_array_cache : std_logic_vector(cBoxcarBx - 1 downto 0) := (others => '0');
        signal data_in           : std_logic_vector(cFIFOWidth - 1 downto 0);
        
        signal link_uncompressed_stubs : ldata(cBoxcarBx - 1 downto 0) := (others => LWORD_NULL); 
    begin
    
        --==============================--
        pRouteInputData : process(clk_p)
        -- Send the data to the correct cache FIFO, using the Bx field of the stub
        --==============================--
        begin
            if rising_edge(clk_p) then
                wr_en_array_cache <= (others => '0');
                wr_en_array_cache(to_integer(unsigned(links_in(i).data(21 downto 19)))) <= links_in(i).valid;
                data_in <= std_logic_vector(to_unsigned(i, 4)) & links_in(i).data(cFIFOWidth - 1 - 4 downto 0);
            end if;
        end process pRouteInputData;


        --==============================--
        genLinkBxCache : for j in cBoxcarBx - 1 downto 0 generate
        -- Sub-generator to produce the FIFO instances for each Bx on each link
        --==============================--
            signal dout_cache                                                                         : std_logic_vector(cFIFOWidth - 1 downto 0);
            signal rd_en_cache, full_cache, empty_cache, valid_cache, almost_empty_cache              : std_logic                                 := '0';
            signal count_cache                                                                        : std_logic_vector(3 downto 0);
            
            signal wr_en_route, rd_en_route, empty_route, full_route, valid_route, almost_empty_route : std_logic                                 := '0';
            signal din_route, dout_route                                                              : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
            signal count_route                                                                        : std_logic_vector(4 downto 0);
            signal route_reset                                                                        : std_logic                                 := '0';
            signal countdown_iterator                                                                 : unsigned(3 downto 0)                      := (others => '0');

        begin
            
            --==============================--
            pTransferToRoutingFIFO : process(clk_p)
            -- Process to send stubs from the cache FIFO to the routing FIFO, so that the stubs can be processed in the routing stage
            --==============================--
            begin
                if rising_edge(clk_p) then
                    -- If FIFO is not empty at start of packet, enable readout to routing FIFO
                    if packet_start = '1' then
                        if empty_cache = '0' then
                            rd_en_cache <= '1';
                            countdown_iterator <= unsigned(count_cache) - 1;
                        end if;
                    else
                        if countdown_iterator > 0 then
                            countdown_iterator <= countdown_iterator - 1;
                        else
                            countdown_iterator <= countdown_iterator;
                            rd_en_cache <= '0';
                        end if;
                    end if;
                    -- Check for non-empty FIFOs at reset
                    if reset = '1' then 
                        non_empty_fifo_count(N_LINKS*j + i) <= (others => '0');
                    else
                        if route_reset = '1' then
                            if empty_route = '0' then
                                non_empty_fifo_count(N_LINKS*j + i) <= std_logic_vector(unsigned(non_empty_fifo_count(N_LINKS*j + i)) + 1);
                            end if;
                        end if;
                    end if;
                end if;
            end process pTransferToRoutingFIFO;
            
            --==============================--
            pCheckBxValidConditionPerLink : process(clk_p)
            --==============================--
            begin
                if rising_edge(clk_p) then
                    if packet_start = '1' then
                        if unsigned(count_cache) >= 2 then
                            bx_valid_array(N_LINKS*j + i) <= '1';
                        else
                            bx_valid_array(N_LINKS*j + i) <= '0';
                        end if;
                    end if;
                end if;
            end process pCheckBxValidConditionPerLink;
            
            
            -- Enable writing of routing FIFO using the valid flag of the cache FIFO
            wr_en_route <= valid_cache;
            din_route <= dout_cache;
            uncompressed_stubs(cBoxcarBx * i + j).data <= dout_cache;
            uncompressed_stubs(cBoxcarBx * i + j).valid <= valid_cache;
            
            
            --==============================--
            SingleLinkCacheInstance : link_agg_single_link_fifo
            --==============================--
            PORT MAP (
                --- Input Ports ---
                clk          => clk_p,
                srst         => reset,
                din          => data_in,
                wr_en        => wr_en_array_cache(j),
                rd_en        => rd_en_cache,
                --- Output Ports ---
                dout         => dout_cache,
                full         => full_cache,
                overflow     => cache_overflow_array(j)(i),
                empty        => empty_cache,
                valid        => valid_cache,
                almost_empty => almost_empty_cache,
                data_count   => count_cache
            );
            --==============================--
            SingleLinkRoutingFIFOInstance : link_agg_single_link_fifo_fwft
            -- Routing FIFO uses First Word Fall Through (FWFT) to reduce the latency.
            --==============================--
            PORT MAP (
                --- Input Ports ---
                clk        => clk_p,
                srst       => route_reset, -- Reset FIFO for every packet
                din        => din_route,
                wr_en      => wr_en_route,
                rd_en      => rd_en_route_array(j)(i),
                --- Output Ports ---
                dout       => data(j)(i),
                full       => full_route,
                overflow   => route_overflow_array(j)(i),
                empty      => empty_route,
                valid      => valid_route_array(j)(i),
                data_count => count_route
            );

            -- Combine the IPBus 'reset' with the 'packet_start' signal to generate the reset signal for the routing FIFO.
            route_reset <= reset or packet_start;
                
        end generate genLinkBxCache;

    end generate genLinkCache;


    --==============================--
    genAggregatedFIFOs : for i in 7 downto 0 generate
    -- Generate the output FIFO instances, one per Bx, as well as the logic to route the stubs from the link FIFOs to the output.
    --==============================--
        signal pointer                                        : integer                                   := 0;
        signal din                                            : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
        signal wr_en, rd_en, full, empty, almost_empty, valid : std_logic                                 := '0';
        signal dout                                           : std_logic_vector(cFIFOWidth - 1 downto 0) := (others => '0');
        signal data_count                                     : std_logic_vector(5 downto 0);
        
        signal compressed_stub                                : std_logic_vector(63 downto 0)             := (others => '0');
        signal compressed_stub_pointer                        : integer range 0 to 1                      := 0;
        signal previous_valid_row                             : std_logic_vector(N_LINKS - 1 downto 0)    := (others => '0');
        signal all_links_reset                                : std_logic                                 := '0';
        
        signal bx_valid                                       : std_logic                                 := '0';

    begin
        --==============================--
        pIteratePointer : process(clk_p)
        -- Process to iterate of the link FIFOs and route the data to the output FIFO, if data is present on a given link.
        -- The process uses a rotating pointer to repeatedly iterate over all links, taking one stub per iteration, until all link FIFOs are empty.
        --==============================--
            variable bx_valid_variable : std_logic := '0';
        
        begin

            if rising_edge(clk_p) then
                -- Loop pointer over all the links
                if pointer = N_LINKS - 1 then
                    pointer <= 0;
                else
                    pointer <= pointer + 1;
                end if;
                -- Readout value of FIFO which is currently pointed to
                rd_en_route_array(i)          <= (others => '0');
                rd_en_route_array(i)(pointer) <= valid_route_array(i)(pointer);
                previous_valid_row            <= valid_route_array(i);
                
--                uncompressed_stubs(i) <= LWORD_NULL;
                if valid_route_array(i)(pointer) = '1' then
                    if compressed_stub_pointer = 1 then
                        compressed_stub_pointer <= 0;
                         wr_en <= '1';
                    else
                        compressed_stub_pointer <= 1;
                        compressed_stub <= (others => '0');
                        wr_en <= '0';
                    end if;
                    compressed_stub(32*compressed_stub_pointer + 31 downto 32*compressed_stub_pointer) <= compress64bStubTo32b(data(i)(pointer));
--                    uncompressed_stubs(i).valid <= valid_route_array(i)(pointer);
--                    uncompressed_stubs(i).data <= data(i)(pointer);
                elsif valid_route_array(i) = cNullEnableArrayRow then
                    if previous_valid_row /= cNullEnableArrayRow and compressed_stub_pointer = 1 then
                        wr_en <= '1';
                    else
                        compressed_stub_pointer <= 0;
                        wr_en                   <= '0';
                        compressed_stub         <= (others => '0');
                    end if;
                else
                    wr_en <= '0';
                end if;
                
                -- Check if Bx is valid
                bx_valid_variable := '1';
                for j in 0 to N_LINKS - 1 loop
                    bx_valid_variable := bx_valid_variable and bx_valid_array(N_LINKS*i + j);
                end loop;
                bx_valid <= bx_valid_variable;               
                
                -- If counter is at the end of a packet, then start to readout the contents of the aggregated FIFOs
                if counter = 47 then
                    if empty = '0' then
                        rd_en <= '1';
                    end if;
                end if;
                if rd_en = '1' then
                    if almost_empty = '1' then
                        rd_en <= '0';
                    end if;
                end if;

                if valid = '1' then
                    links_out(i).valid <= '1';
                    links_out(i).data  <= dout;
                else
                    links_out(i).valid <= '0';
                    links_out(i).data  <= (others => '0');
                end if;               
                links_out(i).strobe <= '1';

            end if;
            
        end process pIteratePointer;
        
        din <= compressed_stub;
        
        --==============================--
        AllLinksFIFOInstance : link_agg_all_links_fifo
        --==============================--
        PORT MAP (
            --- Input Ports ---
            clk          => clk_p,
            srst         => all_links_reset, -- Reset FIFO on every packet
            din          => din,
            wr_en        => wr_en,
            rd_en        => rd_en,
            --- Output Ports ---
            dout         => dout,
            full         => full,
            overflow     => output_overflow_array(i),
            empty        => empty,
            almost_empty => almost_empty,
            valid        => valid,
            data_count   => data_count
        );

        all_links_reset <= reset or packet_start;

    end generate;

    --==============================--
    pReadoutReset : process(clk_p)
    -- Counter process to generate the 'readout_reset' signal which is used downstream in the Link Combiner.
    --==============================--
    begin
        if rising_edge(clk_p) then
            if packet_start = '1' then
                counter <= 0;
            else
                if counter = 64 then
                    counter <= counter;
                else
                    counter <= counter + 1;
                end if;
            end if;
            -- If counter is at the end of a packet, then start to readout the contents of the aggregated FIFOs
            if counter = 48 then
                readout_reset <= '1';
            else
                readout_reset <= '0';
            end if;
        end if;
    end process pReadoutReset;

    --==============================--
    pCountValidBits : process(clk_p)
    --==============================--
        variable bx_valid_variable : std_logic := '0';
    begin
        if rising_edge(clk_p) then
            if counter = 47 then 
                pBx_counter <= 0;
            else
                if pBx_counter = 8 then
                    pBx_counter <= pBx_counter;
                else
                    pBx_counter <= pBx_counter + 1;
                end if;
            end if;

            if pBx_counter < 8 then
                bx_valid_variable := '1';
                for j in 0 to N_LINKS - 1 loop
                    bx_valid_variable := bx_valid_variable and bx_valid_array(N_LINKS*pBx_counter + j);
                end loop;
            end if;

            -- Increment counter if bx_valid is set high when readout is started
            if reset = '1' then 
                bx_valid_count(0) <= (others => '0');
            else
                if bx_valid_variable = '1' and pBx_counter < 8 then
                    bx_valid_count(0) <= std_logic_vector(unsigned(bx_valid_count(0)) + 1);
                end if;
            end if;
        end if;
    end process pCountValidBits;

    --==============================--
    -- Debug
    --==============================--
    status_bits(0) <= output_overflow_array(0); -- link_aggregator_output_0_overflow;
    status_bits(1) <= output_overflow_array(1); -- link_aggregator_output_1_overflow;
    status_bits(2) <= output_overflow_array(2); -- link_aggregator_output_2_overflow;
    status_bits(3) <= output_overflow_array(3); -- link_aggregator_output_3_overflow;
    status_bits(4) <= output_overflow_array(4); -- link_aggregator_output_4_overflow;
    status_bits(5) <= output_overflow_array(5); -- link_aggregator_output_5_overflow;
    status_bits(6) <= output_overflow_array(6); -- link_aggregator_output_6_overflow;
    status_bits(7) <= output_overflow_array(7); -- link_aggregator_output_7_overflow;
    status_bits(8) <= cache_overflow_array(0)(0); -- link_aggregator_cache_0_0_overflow;
    status_bits(9) <= route_overflow_array(0)(0); -- link_aggregator_route_0_0_overflow;
    non_empty_counts <= non_empty_fifo_count;
    bx_valid_count_out <= bx_valid_count;

end compressed;
